"use strict";
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
var prettify = require('object-inspect');
function mkfault(issue) {
    return { path: [], issue: issue };
}
function mkfaults() {
    var issues = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        issues[_i] = arguments[_i];
    }
    return issues.map(mkfault);
}
var Coerce = /** @class */ (function () {
    function Coerce(run) {
        this.run = run;
    }
    Object.defineProperty(Coerce.prototype, "fallback", {
        get: function () {
            return this.run(undefined)[0];
        },
        enumerable: false,
        configurable: true
    });
    Coerce.prototype.go = function (dirty) {
        var _a = this.run(dirty), clean = _a[0], faults = _a[1];
        if (faults.length > 0) {
            var warning = '[kempt warning] Invalid value encountered';
            warning += "\n\nReceived val = " + prettify(dirty);
            warning += '\n\nIssues:';
            for (var _i = 0, faults_1 = faults; _i < faults_1.length; _i++) {
                var fault = faults_1[_i];
                warning += "\n  - val" + joinPath(fault.path) + " " + fault.issue;
            }
            warning += "\n\nValue patched to: " + prettify(clean);
            console.warn(warning);
        }
        return clean;
        function joinPath(path) {
            return path.map(function (key) { return '1234567890'.includes(key[0]) ? "[" + key + "]" : "." + key; }).join('');
        }
    };
    return Coerce;
}());
function k(test, fallback, fixes) {
    if (fixes === void 0) { fixes = []; }
    var fault = test(fallback);
    if (fault)
        throw Error('Provided fallback is invalid.');
    return new Coerce(function (dirty) {
        var faults = [];
        var issue = test(dirty);
        if (!issue)
            return [dirty, []];
        faults.push(mkfault(issue));
        for (var idx = 0; idx < fixes.length; idx++) {
            var fix = fixes[idx];
            var passed = void 0;
            try {
                passed = fix.test(dirty);
            }
            catch (ex) {
                faults.push(mkfault("Fix #" + idx + " failed (test function crashed)."));
                continue;
            }
            if (!passed)
                continue;
            var clean = void 0;
            try {
                clean = fix.run(dirty);
            }
            catch (ex) {
                faults.push(mkfault("Fix #" + idx + " failed (coerce function crashed)."));
                continue;
            }
            var cleanFault = test(clean);
            if (typeof cleanFault === 'string') {
                faults.push(mkfault("Fix #" + idx + " failed (produced indirtyid dirtyue)."));
                continue;
            }
            return [clean, faults];
        }
        return [fallback, faults];
    });
}
var nully = function (dirty) {
    return dirty !== null && dirty !== undefined && "isn't null or undefined";
};
// includes NaN or infinities
var float = function (dirty) {
    return typeof dirty !== "number" && "isn't a number";
};
var number = function (dirty) {
    return float(dirty)
        || Number.isNaN(dirty) && 'is NaN'
        || !Number.isFinite(dirty) && "isn't finite";
};
var integer = function (dirty) {
    return number(dirty)
        || !Number.isInteger(dirty) && "isn't an integer";
};
var bigint = function (dirty) {
    return !(dirty instanceof BigInt) && "isn't a bigint";
};
var string = function (dirty) {
    return typeof dirty !== 'string' && "isn't a string";
};
var boolean = function (dirty) {
    return typeof dirty !== 'boolean' && "isn't a boolean";
};
function Object_map(obj, mapper) {
    return Object.fromEntries(Object.entries(obj).map(function (_a) {
        var key = _a[0], val = _a[1];
        return mapper(key, val);
    }));
}
function bump(faults, key) {
    return faults.map(function (_a) {
        var path = _a.path, issue = _a.issue;
        return ({ path: __spreadArrays(['' + key], path), issue: issue });
    });
}
function object(shape) {
    if (shape === null || typeof shape !== 'object')
        throw Error('Expected an object');
    for (var _i = 0, _a = Object.values(shape); _i < _a.length; _i++) {
        var coerce_1 = _a[_i];
        if (!(coerce_1 instanceof Coerce))
            throw Error('Expected a Coerce');
    }
    return new Coerce(function (dirty) {
        if (dirty === null || typeof dirty !== 'object') {
            var fallback = Object_map(shape, function (key, coerce) { return [key, coerce.fallback]; });
            return [fallback, mkfaults("isn't an object")];
        }
        var dirtyr = dirty;
        var faults = [];
        if (![Object.prototype, null].includes(Object.getPrototypeOf(dirty)))
            faults.push(mkfault('has a prototype other than Object.prototype or null'));
        faults.push.apply(faults, Object.entries(shape).flatMap(function (_a) {
            var key = _a[0], coerce = _a[1];
            return bump(coerce.run(dirtyr[key])[1], key);
        }));
        faults.push.apply(faults, Object.keys(dirtyr).filter(function (key) { return !(key in shape); }).flatMap(function (key) { return bump(mkfaults('is superfluous'), key); }));
        var result = Object_map(shape, function (key, coerce) { return [key, coerce.run(dirtyr[key])[0]]; });
        return [result, faults];
    });
}
function checkLength(expected, actual) {
    if (actual === expected - 1)
        return [{ path: ["" + actual], issue: 'is missing' }];
    else if (actual < expected)
        return [{ path: [actual + ".." + (expected - 1)], issue: 'are missing' }];
    else if (actual === expected + 1)
        return [{ path: ["" + expected], issue: 'is out of bounds' }];
    else if (actual > expected)
        return [{ path: [expected + ".." + (actual - 1)], issue: 'are out of bounds' }];
    else
        return [];
}
function tuple() {
    var shape = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        shape[_i] = arguments[_i];
    }
    if (!(shape instanceof Array))
        throw Error('Expected Array');
    for (var _a = 0, shape_1 = shape; _a < shape_1.length; _a++) {
        var item = shape_1[_a];
        if (!(item instanceof Coerce))
            throw Error('Expected Coerce');
    }
    return new Coerce(function (dirty) {
        if (!(dirty instanceof Array)) {
            var fallback = shape.map(function (coerce) { return coerce.fallback; });
            return [fallback, mkfaults("isn't an array")];
        }
        var faults = [];
        faults.push.apply(faults, checkLength(shape.length, dirty.length));
        faults.push.apply(faults, shape
            .slice(Math.min(shape.length, dirty.length)) // Don't repeat out-of-bounds faults
            .flatMap(function (coerce, idx) { return bump(coerce.run(dirty[idx])[1], idx); }));
        var result = shape.map(function (coerce, idx) { return coerce.run(dirty[idx])[0]; });
        return [result, faults];
    });
}
function vector(coerce, length) {
    if (!(coerce instanceof Coerce))
        throw Error('Expected Coerce');
    return new Coerce(function (dirty) {
        if (!(dirty instanceof Array)) {
            var fallback = Array(length).fill(null).map(function () { return coerce.fallback; });
            return [fallback, mkfaults("isn't an array")];
        }
        var faults = dirty.flatMap(function (elem, idx) { return bump(coerce.run(dirty[idx])[1], idx); });
        faults.push.apply(faults, checkLength(length, dirty.length));
        var result = dirty.map(function (elem) { return coerce.run(elem)[0]; });
        return [result, faults];
    });
}
function list(coerce) {
    if (!(coerce instanceof Coerce))
        throw Error('Expected Coerce');
    return new Coerce(function (dirty) {
        if (!(dirty instanceof Array)) {
            return [[], mkfaults("isn't an array")];
        }
        var faults = dirty.flatMap(function (elem, idx) { return bump(coerce.run(dirty[idx])[1], idx); });
        var result = dirty.map(function (elem, idx) { return coerce.run(dirty[idx])[0]; });
        return [result, faults];
    });
}
var coerce = object({
    latex: k(string, ''),
    inline: k(boolean, false),
});
var result = coerce.go({ x: Number.EPSILON, y: 3 });
